#include <memory>
#include <vector>
#include <string>
#include <fstream>
#include <iostream>
#include <functional>
#include <string_view>
#include <unordered_set>
#include <unordered_map>
#include <initializer_list>

// hint: use valgrind, there seems to be a memory error

struct StringPool
{
    std::string_view ref( const std::string &str ) {
        return std::string_view( *_storage.insert( str ).first );
    }

  private:
    std::unordered_set< std::string > _storage;
};

struct FileCleaner
{
    void add( std::string_view file, std::function< void( std::string_view ) > cleanup ) {
        auto [ it, inserted ] = _cleanups.try_emplace( file, cleanup );
        if ( !inserted ) {
            // chain cleanups, newer runs first
            auto original = it->second;
            _cleanups.emplace( file, [=]( std::string_view ref ) {
                    cleanup( ref );
                    original( ref );
                } );
        }
    }

    void drop( std::string_view file ) {
        auto it = _cleanups.find( file );
        if ( it != _cleanups.end() ) {
            it->second( it->first );
            _cleanups.erase( it );
        }
    }

    ~FileCleaner() {
        for ( auto & [ file, cleanup ] : _cleanups )
            cleanup( file );
    }

  private:
    std::unordered_map< std::string_view, std::function< void( std::string_view ) > > _cleanups;
};

bool exists( const char *file ) {
    std::ifstream s( file );
    return s.good();
}

struct TmpFiles {

    explicit TmpFiles( std::initializer_list< std::string > initial ) :
        _cleaner( std::make_unique< FileCleaner >() ),
        _pool( std::make_unique< StringPool >() )
    {
        for ( const auto &f : initial ) {
            if ( exists( f.c_str() ) ) {
                throw std::runtime_error( "File exists" );
            }
            _cleaner->add( _pool->ref( f ), []( std::string_view ref ) {
                    std::cerr << "deleting initial " << ref.data() << std::endl;
                } );
        }
    }

    ~TmpFiles() {
        // ordering is important
        _cleaner = nullptr;
        _pool = nullptr;
    }

    void make( const std::string &file ) {
        std::cerr << "creating " << file << std::endl;
        _cleaner->add( _pool->ref( file ), []( std::string_view ref ) {
                std::cerr << "deleting " << ref.data() << std::endl;
            } );
    }

  private:
    std::unique_ptr< FileCleaner > _cleaner;
    std::unique_ptr< StringPool > _pool;
};

int main() {
    // this should print "deleting some_file_with_very_long_file_name.txt" but does not!
    try {
        TmpFiles files( { "some_file_with_very_long_file_name.txt", "problem7.cpp" } );
        files.make( "bla" );
        files.make( "blabla" );
    }
    // make sure destructors are run
    catch ( std::exception &ex ) {
        std::cerr << "FATAL: " << ex.what() << std::endl;
    }
}
